/*
 *  linux/fs/lids.c
 *
 *  Author : Xie Hua Gang , (xie@gnuchina.org) 
 *
 *	This file contain the lids security options.
 *
 *	the security option define two states of the kernel.
 *		1.  normality , boot with LILO parameter of "security=0"
 *		    this state is the same as the formal linux.
 *		2.  security , boot with kenerl compiler with  .
 *		    this state is the security state with protection
 *		    the defined file/diretory in  "/etc/ids.conf" to 
 *		    be modified ,such as delete ,chmod , chown ,etc.
 *
 *		    no one can write to the MBR of the device protected.
 *		    for the APPEND ONLY file your protected , it can only
 *		    be open as "APPEND ONLY".
 *
 *	EVEN the Root can not modify the files.
 *		
 *    	with this tools, you can protection your linux box from changing
 *	the important system files, such as "/bin/login" , "/bin/lilo.conf",
 *	"/etc/passwd", etc. 
 *
 */

/*
 *  Changes:
 *	
 *	[Oct 14 1999, Xie Hua Gang] initial creation 
 *	[Oct 17 1999, Xie Hua Gang] Added to do_truncate,do_link(). 
 *	[Oct 27 1999, Xie Hua Gang] Added Append Only ,Device MBR protection. 
 *      [Dec 05 1999, Philippe Biondi] New design. Added hidding process support. Added /dev/[k]mem protection.
 *      [Dec 08 1999, Xie Hua Gang] Added firewall rules protection.
 *      [Dec 09 1999, Philippe Biondi] Added ptrace protection. Permission flags for /dev/mem.
 *      [Dec 17 1999, Philippe Biondi] Added raw disk access, ioperm, iopl protection.
 *      [Jan 08 2000, Philippe Biondi] Use of capabilities bounding set.
 *
 * 	[May 07 2000, Xie Huagang] Add ACL in kernel, reference monitor.
 */

#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/quotaops.h>

#include <asm/uaccess.h>
#include <asm/unaligned.h>
#include <asm/semaphore.h>
#include <asm/page.h>
#include <asm/pgtable.h>

#include <asm/namei.h>

#include <linux/utime.h>
#include <linux/file.h>
#include <linux/fs.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/config.h>
#include <linux/fs.h>
#include <sys/syscall.h>
#include <malloc.h>
#include <linux/tty.h>
#include <linux/sysctl.h>
#include <linux/rmd160.h>

#include <linux/lids.h>

/***********************************************************************
 *********************************************************************** 
 *
 *  Here are some consts and type definitions
 *
 ***********************************************************************
 ***********************************************************************/

#define LIDS_CONF_FILE	"/etc/lids/lids.conf"	/* the configuration file */
#define LIDS_CONF_DIR "/etc/lids"
#define LIDS_ADM_PATH "/sbin/lidsadm"              /* the adminstration tool */
#define LIDS_PW_FILE "/etc/lids/lids.pw"
#define LIDS_PW_LEN  32
#define LIDS_CAP_FILE "/etc/lids/lids.cap"

typedef struct
{
	int last_secure;
	int last_s_acl;
	int last_o_acl;
	unsigned long search_secure;
	unsigned long search_s_acl;
	unsigned long fastguess[2048];
	struct secure_ino secure[CONFIG_LIDS_MAX_INODE];
	struct lids_sys_acl s_acl[CONFIG_LIDS_MAX_SACL];
	struct lids_acl o_acl[CONFIG_LIDS_MAX_OACL];
} _lids_data_t;

/***********************************************************************
 ***********************************************************************
 *
 * General variables
 *
 ***********************************************************************
 ***********************************************************************/

int lids_load=0;        /* it is raised to 1 when kernel boot */
int lids_local_on=1;
lids_flags_t lids_flags=0;
int lids_local_pid=0;
unsigned long lids_current=0;

int lids_first_time=1;

unsigned long lids_cap_val=0; /* store the capability value in lids.cap */

#ifdef CONFIG_LIDS_ALLOW_SWITCH
static char lids_pw[160];
#endif

#ifndef CONFIG_LIDS_ALLOW_ANY_PROG_SWITCH
static struct allowed_ino lidsadm;
#endif

static _lids_data_t lids_data[2]={{0,0},{0,0}};

#ifdef CONFIG_LIDS_RELOAD_CONF
static int lids_lock_init=0;
static spinlock_t lids_lock;
#endif

static unsigned long lids_bittab[32]=
{
	0x00000001,0x00000002,0x00000004,0x00000008,
	0x00000010,0x00000020,0x00000040,0x00000080,
	0x00000100,0x00000200,0x00000400,0x00000800,
	0x00001000,0x00002000,0x00004000,0x00008000,
	0x00010000,0x00020000,0x00040000,0x00080000,
	0x00100000,0x00200000,0x00400000,0x00800000,
	0x01000000,0x02000000,0x04000000,0x08000000,
	0x10000000,0x20000000,0x40000000,0x80000000
};

/***********************************************************************
 ***********************************************************************
 *
 * LIDS protection management
 *
 ***********************************************************************
 ***********************************************************************/

int lids_local_off(void) 
{
	struct task_struct *t;
	read_lock(&tasklist_lock);
	t=current;
	while (t && (t->pid > 1)) {
		if (t->pid == lids_local_pid) {
			read_unlock(&tasklist_lock);
			return 1;
		}
		t=t->p_pptr;
	}
	read_unlock(&tasklist_lock);
	return 0;
}

/*
 *
 *	LIDS ACL Function 
 *
 */ 

/*
 *      used when execed, load the lids_cap into task_struct.
 */

static __inline__ struct lids_sys_acl * lids_do_search_acl(unsigned long int ino,kdev_t dev,_lids_data_t *data)
{
	long i;
	long j;

	/* ok, in case of e.g. 256 entries we would require 256 comparisons
	   for a linear search with no match, the following reduces this
	   to 8 comparisons (table is sorted!) */

	for(j=i=data->search_s_acl;;j>>=1)
	{
		if(i>=data->last_s_acl)i-=j;
		else if(data->s_acl[i].dev<dev)i+=j;
		else if(data->s_acl[i].dev>dev)i-=j;
		else if(data->s_acl[i].ino<ino)i+=j;
		else if(data->s_acl[i].ino>ino)i-=j;
		else return &(data->s_acl[i]);
		if(!j||i<0)return NULL;
	}
}

struct lids_sys_acl * lids_search_acl(unsigned long int ino,kdev_t dev,unsigned long lids_curr)
{
	return lids_do_search_acl(ino,dev,&lids_data[lids_curr&1]);
}

/*
 *	used in before access file(lids_check_base),
 *	check if the program have permission to access the files.
 *	if it has, then dont access the files.
 *
 *	we must sure that current->lids_cap is not NULL when call.
 */ 
static int lids_check_acl(struct dentry *base,int type,unsigned long lids_curr,int check_domain)
{
	struct lids_acl *acl;
	struct dentry *parent;
	struct inode *inode;
	
#ifdef NO_CONFIG_LIDS_RELOAD_CONF
	if(lids_curr!=current->lids_current && current->lids_sys_acl && current->p_pptr)
	{
		struct lids_sys_acl *that_sys_acl;
		struct task_struct *parent=current->p_pptr;
		struct lids_sys_acl *p_sys_acl;
		/* minimal race condition here: in the worst case the
		following code is executed a second time but this
		is better than a global or per process spinlock */
		spin_lock(&lids_lock);
		current->lids_current=lids_current;
		that_sys_acl = lids_do_search_acl(current->lids_ino,current->lids_dev,&lids_data[lids_curr&1]); 
		if(that_sys_acl)
	  		memcpy(current->lids_sys_acl,that_sys_acl,sizeof(struct lids_sys_acl));
		else
	  		memcpy(current->lids_sys_acl,0,sizeof(struct lids_sys_acl));
		/* get the parent's acls and capabilities */
		if((p_sys_acl=parent->lids_sys_acl)!=NULL) {
			current->lids_sys_acl->flags |= (p_sys_acl->flags&p_sys_acl->inherit_flags);
			current->lids_sys_acl->inherit_flags |=(p_sys_acl->inherit_flags);	
			/* IMCOMPLETE: reinsert the parent's inherit ACLs */
		}
		spin_unlock(&lids_lock);
	}
#endif
	if( base == NULL || current->lids_sys_acl==NULL)
		return -EROFS; 
	/*
	LIDS_DBG(" %d -- %ld,%d\n",type,base->d_inode->i_ino,base->d_inode->i_dev);
*/
	inode  = base->d_inode;
	parent = base->d_parent;
	
	if(parent == NULL)
		return -EROFS;
	
	if ( inode == NULL  ) {
		LIDS_DBG("inode == NULL\n");
		inode = parent->d_inode;
		parent = parent->d_parent;
	}
	
	do {
		LIDS_DBG("checking for (%p,%p) %ld,%d\n",base,parent,inode->i_ino,inode->i_dev);
		if(check_domain)
			acl=current->lids_sys_acl->lids_domain;
		else
			acl = current->lids_sys_acl->lids_acl;
		while(acl != NULL) {
			LIDS_DBG("checking loop for %ld,%d vs %ld,%d\n",acl->ino,acl->dev,inode->i_ino,inode->i_dev);
			if(acl->ino==inode->i_ino && acl->dev==inode->i_dev) {
				return type&acl->type?0:-EROFS;
			}
			acl = acl->next;
		}
		if ( inode == parent->d_inode) {
			LIDS_DBG("inode == parent->d_inode,return -1\n");
			break;
		}
		inode = parent->d_inode;
	} while( ((parent = parent->d_parent ) != NULL));

	return -1;	
}
/*
 *	lids_set_acls, this_sys_acl must be NOT NULL.
 */ 
int lids_set_acls(struct lids_sys_acl *this_sys_acl)
{

	struct lids_acl *t_acl,*this_acl,*last_acl;
	int	i;
	
	LIDS_DBG("current->cap_effective = %x\n",current->cap_effective);

	/* current->p_pptr,get parent's inherit_flags. 
	 * get the inherit from it's direct parent
	 * if the inherit is not NULL, inherit the capability 
	 * or clear the capability */
	
	current->lids_cap = 0;

	/* exam the inherit first from current->lids_sys_acl*/
	/* 1. CAP inherit */
	if(current->lids_sys_acl->forked) {

	if(current->lids_sys_acl->flags !=0 ) {
		current->lids_sys_acl->flags = 0;
		/* reset the cap_inherit */
		for(i=0;i<32;i++) {
			if(current->lids_sys_acl->cap_inherit[i] !=0) {
				set_bit(i,&current->lids_sys_acl->flags);
			}
			if(current->lids_sys_acl->cap_inherit[i] > 0) {
				current->lids_sys_acl->cap_inherit[i]--;
			}
		}
	}
	/* 2. File Acls inherit */
  	this_acl = current->lids_sys_acl->lids_acl;
	last_acl = NULL;	
	i=0; 
	while(this_acl) {
		if(this_acl->inherit == 0) {
			/*delete the no inheritabe acls */
			if(i == 0)
				current->lids_sys_acl->lids_acl=this_acl->next;
			else 
				last_acl->next = this_acl->next;
			kfree(this_acl);
		}
		else if(this_acl->inherit > 0)
			this_acl->inherit--;
		if(this_acl->inherit !=0) {
			last_acl = this_acl;
			i++;
		}
		this_acl =this_acl->next;
	}
	current->lids_sys_acl->forked=0;
	}
		
        /* Set its own ACLs and CAP */
	if(this_sys_acl) {
		int	i,inherit;
		int 	this_inherit;

		current->lids_sys_acl->flags |= this_sys_acl->flags;
		
		/* set the cap_inherit array*/
		for( i=0;i<32;i++) {
			inherit = current->lids_sys_acl->cap_inherit[i];
			this_inherit = this_sys_acl->cap_inherit[i];
			if((inherit!=0 && inherit<this_inherit) || inherit==0) 
				current->lids_sys_acl->cap_inherit[i] = this_inherit;
			if(current->lids_sys_acl->cap_inherit[i]!=0) {
				set_bit(i,&current->lids_sys_acl->flags);
			}
		}
		/* FIXME:The domain change to current running process's domain*/
		current->lids_sys_acl->lids_domain = this_sys_acl->lids_domain;
		/* get the lids acl from the this proceeds's ACL-list */
 		t_acl=this_sys_acl->lids_acl;

		last_acl = current->lids_sys_acl->lids_acl;
		while(t_acl) {
			this_acl =  kmalloc(sizeof(struct lids_acl), GFP_KERNEL);
			if(!this_acl) return -1;
			memcpy(this_acl,t_acl,sizeof(struct lids_acl));
			this_acl->next=last_acl;
			last_acl = this_acl;
			t_acl = t_acl->next;
		}
		current->lids_sys_acl->lids_acl = last_acl;
	}

        /* from the inherit flags, set current flags
	 * Do you consider the INIT --> all the children, FIXME
	 */
        current->lids_cap = current->lids_sys_acl->flags;
	LIDS_DBG("set_acls: current->lids_cap =%x\n", current->lids_cap);

	return 0;
}
/*
 *	free alloc current->lids_sys_acl when do_exit()
 *	FIXME: problem when releasing the kmalloc() ..
 */
void exit_lids(void)
{
	struct lids_sys_acl * this_sys_acl = current->lids_sys_acl;
	struct lids_acl	*this_acl,*next_acl;

	if (this_sys_acl) {
		this_acl = this_sys_acl->lids_acl;
		while(this_acl) {
			next_acl = this_acl->next;
			kfree(this_acl);
			this_acl = next_acl;
		}
		LIDS_DBG("pid=%d begin to free %p\n",current->pid,this_sys_acl);
		kfree(this_sys_acl);
	}
}
/*
 *	lids_add_acl 
 *	add the acl into system.
 */ 
static void lids_add_acl(unsigned long int s_ino ,kdev_t s_dev , unsigned long int o_ino,kdev_t o_dev, int type,int inherit, _lids_data_t *data) 
{
        struct lids_sys_acl *this_sys_acl;
        struct lids_acl *lids_acl,*acl;
	int i;

	for(this_sys_acl=NULL,i=0;i<data->last_s_acl&&!this_sys_acl;i++)
	    if(data->s_acl[i].ino==s_ino&&data->s_acl[i].dev==s_dev)
		this_sys_acl=&(data->s_acl[i]);

        if(!this_sys_acl)
	{
		if(data->last_s_acl==CONFIG_LIDS_MAX_SACL)return;
		if((int)(o_ino)!=-1&&data->last_o_acl==CONFIG_LIDS_MAX_OACL)
			return;

	        LIDS_DBG("lids_add_s_acl,%ld,%d\n",s_ino,s_dev);
 
        	this_sys_acl=&(data->s_acl[(data->last_s_acl)++]);
	        this_sys_acl->ino = s_ino;
	        this_sys_acl->dev = s_dev;
		this_sys_acl->flags = 0;
		memset(this_sys_acl->cap_inherit,'\0',32*sizeof(int *));
		this_sys_acl->lids_acl = NULL;
		this_sys_acl->lids_domain = NULL;
	}

	if((int)(o_ino)==-1)
	{
		/* flags indicate that this proceed has capability 
		 * to do someting ,inherit_flags dicide that if the capability
		 * can be inherit to its children */
		set_bit(o_dev,&this_sys_acl->flags);
		this_sys_acl->cap_inherit[o_dev] = inherit;
		return;
	}

	if(data->last_o_acl==CONFIG_LIDS_MAX_OACL)return;

	lids_acl=&(data->o_acl[(data->last_o_acl)++]);
	lids_acl->ino = o_ino;
	lids_acl->dev = o_dev;
	lids_acl->inherit = inherit;
	lids_acl->next = NULL;

	if(type < 0) {
		lids_acl->type = -type;
		if((acl=this_sys_acl->lids_domain)==NULL)
			this_sys_acl->lids_domain=lids_acl;
		else {
			while(acl->next)acl=acl->next;
			acl->next = lids_acl;
		}
	}
	else {
		lids_acl->type = type;
		if((acl=this_sys_acl->lids_acl)==NULL)
			this_sys_acl->lids_acl=lids_acl;
		else {
			while(acl->next)acl=acl->next;
			acl->next = lids_acl;
		}
	}
}

static __inline__ int lids_acl_dev_ino_cmp(struct lids_sys_acl *s1,struct lids_sys_acl *s2)
{
	if(s1->dev<s2->dev)return -1;
	if(s1->dev>s2->dev)return 1;
	if(s1->ino<s2->ino)return -1;
	if(s1->ino>s2->ino)return 1;
	return 0;	/* Aieeeh, duplicate entry */
}

static int lids_sort_acls(_lids_data_t *data)
{
	int i;
	int working;
	struct lids_sys_acl tmp;

	/* sort source acl entries by device and inode */

	do
	{
	    working=0;
	    for(i=0;i<data->last_s_acl-1;i++)
	      switch(lids_acl_dev_ino_cmp(&(data->s_acl[i]),
		&(data->s_acl[i+1])))
	    {
		case 0:	return 0;
		case 1: tmp=data->s_acl[i];
			data->s_acl[i]=data->s_acl[i+1];
			data->s_acl[i+1]=tmp;
			working=1;
	    }
	    for(i=data->last_s_acl-1;i>0;i--)
	      switch(lids_acl_dev_ino_cmp(&(data->s_acl[i-1]),
		&(data->s_acl[i])))
	    {
		case 0:	return 0;
		case 1: tmp=data->s_acl[i-1];
			data->s_acl[i-1]=data->s_acl[i];
			data->s_acl[i]=tmp;
			working=1;
	    }
	} while(working);

	/* set up base index for fast search: no, we can't use >0x20000000
	   here but anyway it'll take some years for computers to
	   allow for the memory required for this size :-) */

	data->search_s_acl=0x20000000;
	while(data->last_s_acl<=data->search_s_acl&&data->search_s_acl)
		data->search_s_acl>>=1;

	return 1;
}

/***********************************************************************
 *
 *	lids_search_inode()	
 *
 *	check the inode in a arr_ino array.
 *	may be use fast search in this array.
 *	xhg, 1999/10/11.
 *
 */

static __inline__ int lids_search_inode(unsigned long ino,kdev_t dev,_lids_data_t *data) 
{
	long i=(ino^dev)&0xffff;
	long j;

	/* when the 'hash' bit is not set we for sure don't have
	   a matching entry, so just signal 'nothing found' */

	if(!(data->fastguess[i>>5]&lids_bittab[i&31]))return -1;

	/* ok, in case of e.g. 256 entries we would require 256 comparisons
	   for a linear search with no match, the following reduces this
	   to 8 comparisons (table is sorted!) */

	for(j=i=data->search_secure;;j>>=1)
	{
		if(i>=data->last_secure)i-=j;
		else if(data->secure[i].dev<dev)i+=j;
		else if(data->secure[i].dev>dev)i-=j;
		else if(data->secure[i].ino<ino)i+=j;
		else if(data->secure[i].ino>ino)i-=j;
		else
		{
			return data->secure[i].type;
		}
		if(!j||i<0)return -1;
	}
}

int lids_check_hidden_inode(unsigned long ino,kdev_t dev)
{
	if(lids_search_inode(ino,dev,&lids_data[lids_current&1])==LIDS_DENY)
		return -1;
	return 0;
}

/***********************************************************************
 *
 *lids_check_base();
 *	
 *	check if the base have been protected by the IDS system.
 *	use the base->d_parent 
 *
 */

/*
 *	lids_check_base 
 *
 *	check if the requried access can be permitted
 */ 

int lids_check_base(struct dentry *base, int flag)
{
	struct inode *ino;
	struct dentry *dentry=base;
	int retval=0;
	unsigned long lids_curr=lids_current;

	/* check if the dentry is in the domain */
	while(dentry) {
		if((ino=dentry->d_inode)!=NULL) {  
			if(current->lids_sys_acl && current->lids_sys_acl->lids_domain)
				retval = lids_check_acl(base,flag,lids_curr,1);
			else
				break;
			if(retval == 0)
				break;
			return -EPERM;
		}
		if (dentry==dentry->d_parent) {
			/* do not in this exec domain */
			return -EPERM;
		}
		dentry=dentry->d_parent;
	}
	
	while (dentry) {
		if((ino=dentry->d_inode)!=NULL)
		    if((retval=lids_search_inode(ino->i_ino,ino->i_dev,
			&lids_data[lids_curr&1])) >= 0)
		{
			return (retval & flag) ? 0:
				lids_check_acl(base,flag,lids_curr,0);
		}
		if (dentry==dentry->d_parent)
			return 0;
		dentry=dentry->d_parent;
	}
	return 0;
}

/***********************************************************************
 *
 *	lids_add_inode(inode,dev,type)
 *
 *	add the given inode into the arr_ino.
 *	if full return 0
 *	else return the insert position.
 *
 */
static void lids_add_inode(unsigned long inode ,kdev_t dev , int type,_lids_data_t *data) 
{
	int i;

	if (data->last_secure==CONFIG_LIDS_MAX_INODE)return;

	LIDS_DBG("lids_add_inode,%ld,%d,%d\n",inode,dev,type);

	for(i=0;i<data->last_secure;i++)
	    if(data->secure[i].ino==inode&&data->secure[i].dev==dev)
		return;

	data->secure[data->last_secure].ino = inode;
	data->secure[data->last_secure].dev = dev;
	data->secure[data->last_secure].type = type;

	(data->last_secure)++;

	i=(dev^inode)&0xffff;
	data->fastguess[i>>5]|=lids_bittab[i&31];
}

static __inline__ int lids_inode_dev_ino_cmp(struct secure_ino *s1,struct secure_ino *s2)
{
	if(s1->dev<s2->dev)return -1;
	if(s1->dev>s2->dev)return 1;
	if(s1->ino<s2->ino)return -1;
	if(s1->ino>s2->ino)return 1;
	return 0;	/* Aieeeh, duplicate entry */
}

static int lids_sort_inodes(_lids_data_t *data)
{
	int i;
	int working;
	struct secure_ino tmp;

	/* sort inode entries by device and inode */

	do {
	    working=0;
	    for(i=0;i<data->last_secure-1;i++)
	      switch(lids_inode_dev_ino_cmp(&(data->secure[i]),
		&(data->secure[i+1]))) {
		case 0:	return 0;
		case 1: tmp=data->secure[i];
			data->secure[i]=data->secure[i+1];
			data->secure[i+1]=tmp;
			working=1;
	      } 
	    for(i=data->last_secure-1;i>0;i--)
	      switch(lids_inode_dev_ino_cmp(&(data->secure[i-1]),
		&(data->secure[i]))) {
		case 0:	return 0;
		case 1: tmp=data->secure[i-1];
			data->secure[i-1]=data->secure[i];
			data->secure[i]=tmp;
			working=1;
	      }
	} while(working);

	/* set up base index for fast search: no, we can't use >0x20000000
	   here but anyway it'll take some years for computers to
	   allow for the memory required for this size :-) */

	data->search_secure=0x20000000;
	while(data->last_secure<=data->search_secure&&data->search_secure)
		data->search_secure>>=1;

	return 1;
}

/**************************************************************************/

/*
 *	lids_get_info(char *buffer);
 *
 */ 
static int lids_get_info(char *buffer, unsigned long int *ino,kdev_t *dev)
{
	int	error = -1;
	char  	ino_str[64],dev_str[64];
	char 	*q,*p;
	struct 	dentry *d_file;
	int 	is_special = 0;

	/* get the dentry of the file */
	LIDS_DBG("lids_get_info : [%s] \n",buffer);
	/* it is a special type, ino=-1*/
	if(*buffer == '-' ) {
		is_special = 1;
	}
	memset(ino_str,'\0',64);
	memset(dev_str,'\0',64);

	p = memscan(buffer,':',strlen(buffer));
	if(((unsigned long)(p-buffer))==strlen(buffer))goto exit;
	if(((unsigned long)(p-buffer))>63)goto exit;
	memcpy(ino_str,buffer,p-buffer);

	p++;
	q = memscan(p,':',strlen(p));
	if(((unsigned long)(q-p))==strlen(p))goto exit;
	if(((unsigned long)(q-p))>63)goto exit;
	memcpy(dev_str,p,q-p);

	error = 0;
	if(is_special) {
		*dev = (kdev_t)simple_strtoul(dev_str,0,0);
		*ino = -1;
		goto exit;
	}
	q++;
	d_file = lookup_dentry(q, NULL, 0);

	if (IS_ERR(d_file) || (!d_file) || !(d_file->d_inode)) {
		*ino = simple_strtoul(ino_str,0,0);
		*dev = (kdev_t)simple_strtoul(dev_str,0,0);
	}
	else {
		if ( d_file->d_mounts == d_file ) {
			LIDS_DBG("OK,Mount point found!\n");
			*ino = simple_strtoul(ino_str,0,0);
			*dev = (kdev_t)simple_strtoul(dev_str,0,0);
		}
		else {
			*ino=simple_strtoul(ino_str,0,0);
			*dev=(kdev_t)simple_strtoul(dev_str,0,0) ;
			if ((*ino != d_file->d_inode->i_ino) ||
			    (*dev != d_file->d_inode->i_dev)) {
				LIDS_DBG("now=%p,mnt=%p\n",d_file,d_file->d_mounts);
				printk("LIDS: [%s] dev/inode in lids.conf seems wrong\n",q);
			}
		}
		dput(d_file);
	}
 exit:	
	LIDS_DBG("lids_get_info : return %d\n",error);
	return error;
}
/***********************************************************************
 *
 *	lids_init_add_file ( char * filename)
 *
 *	read filename from read_file() 
 *
 *	method : get the dentry from the filename.
 *		 check if its parent is in the err_ino.
 *		 if not, add the inode.
 *		 
 *	the format is s_ino:s_dev:s_file:type:o_ino:o_dev:o_file
 */

static int lids_init_add_file ( char *buffer,_lids_data_t *data)
{
	char *p,*q;
	int error = -1;
	int is_default_rule = 0;
	int type,inherit;
	unsigned long int s_ino,o_ino;
	kdev_t	s_dev,o_dev;
	
	p = memscan(buffer,':',strlen(buffer));
	if(((unsigned long)(p-buffer))==strlen(buffer))goto exit;

	p++;
	q = memscan(p,':',strlen(p));
	if(((unsigned long)(q-p))==strlen(p))goto exit;

	q++;
	if(*q == ':' ) {
		is_default_rule = 1;
		p = q;
	}
	else {
		p = memscan(q,':',strlen(q));
		if(((unsigned long)(p-q))==strlen(q))goto exit;
	}

	*p = '\0';
	p++;
	q = memscan(p,':',strlen(p));
	*q = '\0';
	q++;
	type =simple_strtol(p,0,0);

	p = memscan(q,':',strlen(q));
	*p = '\0';
	inherit = simple_strtol(q,0,0);
	p++;

	/* q point to the object's ino,dev,filename */
	if(is_default_rule) {
		s_ino = s_dev = 0;
	}	
	else {
		if ( lids_get_info(buffer,&s_ino,&s_dev) < 0 )
			goto exit;
	}
	if ( lids_get_info(p,&o_ino,&o_dev) < 0 )
		goto exit ;

	if(is_default_rule)
		lids_add_inode(o_ino,o_dev,type,data);
	else {
		lids_add_acl(s_ino,s_dev,o_ino,o_dev,type,inherit,data);
	}
	error = 0;
exit:
	return error;
}

/***********************************************************************
 *
 *	add a file into the secure list.
 *
 */

static int init_add_file(char *filename,_lids_data_t *data) 
{
	struct dentry *dentry;
	int	error=0;


	dentry = lookup_dentry(filename, NULL, 0);

     	if (IS_ERR(dentry)) return -1;
	if (!dentry) return -2;

	if (!(dentry->d_inode))error = -2;
	else lids_add_inode(dentry->d_inode->i_ino,dentry->d_inode->i_dev,
		LIDS_DENY,data);

	dput(dentry);
	LIDS_DBG("init_add_file : return %d\n",error);
	return error;
}
/*
 * read the capability value into lids_cap 
 * format [+|-]number:cap_name
 */
int lids_cap_init(char *buffer)
{
	char *p;
	int flag;
	int error = 0;
	error = -1;	
	p = memscan(buffer,':',strlen(buffer));
	if(((unsigned long)(p-buffer))==strlen(buffer))goto exit;
	
	*p = '\0';
	flag = simple_strtoul(buffer+1,0,0);
	error = 0;
	if(buffer[0] == '+')
		cap_raise(lids_cap_val,flag);
	else if(buffer[0] == '-' )
	       	cap_lower(lids_cap_val,flag);
	else 	
		error = -2;
exit:
	return error;
}

/*
 * lids read capability from /etc/lids/lids.cap 
 */

int lids_read_cap(void)
{
 	struct file     *filp;
        char    buffer[1024], *p,*q;
        mm_segment_t    oldfs;
        int     bytes;
	int 	error =0;
	int	start=0,finished=0;

        filp = filp_open(LIDS_CAP_FILE,O_RDONLY,O_RDONLY);
        if (IS_ERR(filp)||(filp==NULL)) {
	   	error = -1 ;
		printk("LIDS: Checking capability file %s error,does it exit?\n",LIDS_CONF_FILE);
		/* FIXME: if (lids_load) goto err_panic;  */
		return error;
	}

        if (filp->f_op->read==NULL) {
		fput(filp);
	    	error = -3 ;
		printk("LIDS: The file %s can not be read\n",LIDS_CONF_FILE);
		/*
            	if (lids_load) goto err_panic ; 
		*/
		return error;
	}
	while ( !finished ) {
      /* Now read 1024 bytes into buffer */
       		filp->f_pos = start;
       		oldfs = get_fs();
       		set_fs(KERNEL_DS);
       		bytes = filp->f_op->read(filp,buffer,1024,&filp->f_pos);
       		set_fs(oldfs);

       		q = buffer;
		if(bytes<1024){
			finished=1;
			q[bytes++]='\n';
		}
			 
       /* analyze the buffer */
		while(bytes&&(p=memscan(q,'\n',bytes))!=q+bytes)
		{
			*p++='\0';
			start+=(int)(p-q);
			bytes-=(int)(p-q);
			while(*q=='\r')q++; /* hmmm... */
			if(*q == '+' || *q == '-') if ( lids_cap_init(q) < 0)
			{
				error = -5;
				printk("LIDS: error in adding [%s] to the kernel\n",q);
				break;
			}
			q=p;
		}
                if(error)break;

                if(bytes==1024)
                {
                    printk("LIDS: Line too long in file %s\n",LIDS_CAP_FILE);
                    error = -4; /* if the line contains no '\n' */
                    break;
                }
	}

	LIDS_DBG("lids_read_cap:lids_cap_val = %x\n",lids_cap_val);
       /* Close the file */
        fput(filp);
	return error;	
}
/*
 *  read the lids password generated by "lidsadm -P" into kernel 
 *  if nessary 
 */ 
#ifdef CONFIG_LIDS_ALLOW_SWITCH
int lids_read_pw(void)
{
 	struct file     *filp;
        char    buffer[LIDS_PW_LEN];
        mm_segment_t    oldfs;
        int     bytes;
	int 	error =0;

        filp = filp_open(LIDS_PW_FILE,O_RDONLY,O_RDONLY);
        if (IS_ERR(filp)||(filp==NULL)) {
	   	error = -1 ;
		printk("LIDS: Checking passwd file %s error,does it exit?\n",LIDS_CONF_FILE);
		/*
		FIXME: if (lids_load) goto err_panic;  
		*/

		return error;
	}

        if (filp->f_op->read==NULL) {
		fput(filp);
	    	error = -3 ;
		printk("LIDS: The file %s can not be read\n",LIDS_CONF_FILE);
		/*
            	if (lids_load) goto err_panic ; 
		*/
		return error;
	}

      /* Now read LIDS_PW_LEN bytes from postion "StartPos" */
       filp->f_pos = 0;
       oldfs = get_fs();
       set_fs(KERNEL_DS);
       bytes = filp->f_op->read(filp,buffer,LIDS_PW_LEN,&filp->f_pos);
       set_fs(oldfs);

	if(bytes<LIDS_PW_LEN)
		return -1;

	memcpy(lids_pw,buffer,LIDS_PW_LEN);

	LIDS_DBG("lids_pw = %s\n",lids_pw);
       /* Close the file */
        fput(filp);
	return error;	
}
#endif

/***********************************************************************
 *
 *	lids_init
 *
 *	initialize the vfs security system. read the config file .
 *	add the inode to the files.
 *
 */

int lids_init(void)
{
 	struct file     *filp;
        char    buffer[1024], *p,*q;
        mm_segment_t    oldfs;
        int     bytes;
	int	start = 0;
	int	finished=0;
	int	error = 0;
	struct dentry *dentry;
	_lids_data_t *data=&lids_data[(lids_current&1)^1];

	/* Get lidsadm dev/inode */
	
	LIDS_DBG("into lids_init_..\n");
#ifdef CONFIG_LIDS_PORT_SCAN_DETECTOR
	lids_port_scanner_detector_init();
#endif
#ifdef CONFIG_LIDS_RELOAD_CONF
	if(!lids_lock_init)
	{
		spin_lock_init(&lids_lock);
		lids_lock_init=1;
	}
#endif

	/* reset fast guess table */

	memset(data->fastguess,0,sizeof(data->fastguess));

#ifndef CONFIG_LIDS_ALLOW_ANY_PROG_SWITCH
	dentry = lookup_dentry(LIDS_ADM_PATH,NULL,0);
  	if (IS_ERR(dentry) || (!dentry)) {
    		printk("LIDS : " LIDS_ADM_PATH " not found \n");
		lidsadm.ino = 0;
		lidsadm.dev = 0;
  	}
	else {
		if(!dentry->d_inode ||
		   !dentry->d_inode->i_dev ||
		   !dentry->d_inode->i_ino) {
			printk("LIDS: something wrong with " LIDS_ADM_PATH "\n");
			lidsadm.ino = 0;
			lidsadm.dev = 0;
		}
		else {
			lidsadm.ino = dentry->d_inode->i_ino;
			lidsadm.dev = dentry->d_inode->i_dev;
		}
		dput(dentry);
	}
#endif

	/* Now, try to read lids.conf */

	if ( init_add_file(LIDS_CONF_DIR,data) < 0 ) {
		error = -1;
		printk("LIDS: Checking conf file %s error,does it exit?\n",LIDS_CONF_FILE);
		if (lids_load) goto lids_panic;
		return -1;
	}
        filp = filp_open(LIDS_CONF_FILE,O_RDONLY,O_RDONLY);
        if (IS_ERR(filp)||(filp==NULL)) {
	   	error = -2 ;
		printk("LIDS: Checking conf file %s error,does it exit?\n",LIDS_CONF_FILE);
           	if (lids_load) goto lids_panic;  /* Or do something else */
		return -1;
	}

        if (filp->f_op->read==NULL) {
		fput(filp);
	    	error = -3 ;
		printk("LIDS: The file %s can not be read\n",LIDS_CONF_FILE);
            	if (lids_load) goto lids_panic ;  /* File(system) doesn't allow reads */
		return -1;
	}

      /* Now read 1024 bytes from postion "StartPos" */
	while ( !finished ) {
       		filp->f_pos = start;
       		oldfs = get_fs();
       		set_fs(KERNEL_DS);
        	bytes = filp->f_op->read(filp,buffer,1024,&filp->f_pos);
        	set_fs(oldfs);

		q=buffer;

		if(bytes<1024)
		{
			finished=1;
			q[bytes++]='\n';
		}

		while(bytes&&(p=memscan(q,'\n',bytes))!=q+bytes)
		{
		    *p++='\0';
		    LIDS_DBG(" read_file : line=->%s<-\n",q);
		    start+=(int)(p-q);
		    bytes-=(int)(p-q);
		    while(*q=='\r')q++; /* hmmm... */
		    if(*q&&*q!='#')if(lids_init_add_file(q,data)<0)
		    {
			printk("LIDS: error in adding [%s] to the kernel\n",q);
			error = -5;
			break;
		    }
		    q=p;
		}

		if(error)break;

		if(bytes==1024)
		{
		    printk("LIDS: Line too long in file %s\n",LIDS_CONF_FILE);
		    error = -4;	/* if the line contains no '\n' */
		    break;
		}
	}

	memset(buffer,'\0',1024);

       /* Close the file */
        fput(filp);

	if(error)
	{
           	if (lids_load) goto lids_panic;  /* Or do something else */
		return -1;
	}
	if(!lids_sort_inodes(data))
	{
		printk("LIDS: Aieeh, duplicate inode entry in file %s\n",LIDS_CONF_FILE);
		error = -6;
           	if (lids_load) goto lids_panic;  /* Or do something else */
		return -1;
	}

	if(!lids_sort_acls(data))
	{
		printk("LIDS: Aieeh, duplicate source acl entry in file %s\n",LIDS_CONF_FILE);
		error = -7;
           	if (lids_load) goto lids_panic;  /* Or do something else */
		return -1;
	}
#ifdef CONFIG_LIDS_ALLOW_SWITCH
	/* Read the password now */
	if(lids_read_pw()) {
		printk("LIDS: Read password file error\n");
		error = -8;
		goto lids_panic;
	}
#endif
	/* read the cap init from lids.cap into lids_cap_val */
	if(lids_read_cap()) {
		printk("LIDS: Read password file error\n");
		error = -9;
		goto lids_panic;
	}
#ifdef CONFIG_LIDS_SA_THROUGH_NET
	/* read network parameter from lids.net */
	if(lids_read_net()) {
		printk("LIDS: Read mail paramenter error\n");
		error = -10;
		goto lids_panic;
	}
#endif

#if 0 
	printk("read_file : finish=%d ret %d\n",finished,error);
	for(start=0;start<data->last_secure;start++)
	{
		int i;
		struct lids_sys_acl *this_sys_acl;
		struct lids_acl *this_acl;
		this_sys_acl = (struct lids_sys_acl *)&(data->s_acl[start]);

		printk("%ld\t%d\t%d\n",data->secure[start].ino,
				data->secure[start].dev,
				data->secure[start].type);


		/* -------------------*/
                for(i=0;i<32;i++)
			printk("%d-\t",this_sys_acl->cap_inherit[i]);
		this_acl = this_sys_acl->lids_acl;
		while(this_acl!=NULL) {
			printk("----------\nobject:\tthis_acl=%ld,%d,type=%d,inherit=%d\n",this_acl->ino,this_acl->dev,this_acl->type,this_acl->inherit);
			this_acl = this_acl->next;
		}
		/* -------------------*/
	}
#endif

	if (!error ) { 
		LIDS_DBG("written to index %d\n",(lids_current&1)^1);
		printk("LIDS: Statistics: %d objects, %d source ACLS, %d object ACLs,capability = %lx\n",data->last_secure,data->last_s_acl,data->last_o_acl,lids_cap_val);
		lids_current++;
		LIDS_DBG("lids_current=%d index=%d\n",lids_current,lids_current&1);
		return 0;
	}
	else 
lids_panic:
	panic("LIDS: Can not initial the lids system, return code %d, contact Xie Huagang (xie@gnuchina.org)",error);
	return error;
}

/***********************************************************************
 ***********************************************************************
 *
 * Now, the sysctl procedures
 *
 ***********************************************************************
 ***********************************************************************/



/***********************************************************************
 *
 * What is needed to lock init children
 *
 */

pid_t lids_protected_pid[CONFIG_LIDS_MAX_PROTECTED_PID];
int lids_last_pid=0;

static int init_no_kill_pid(void)
{
        struct task_struct * p;
        
        read_lock(&tasklist_lock);
        for_each_task(p) {
                if ( p->pid >1 && p->p_pptr->pid == 1 ) {
                        if( lids_last_pid < CONFIG_LIDS_MAX_PROTECTED_PID) { 
                                lids_protected_pid[lids_last_pid++] = p->pid;
                        }
                }
        }
        read_unlock(&tasklist_lock);
        return 0;
}

/***********************************************************************
 *
 * The one which process flags changes
 *
 * If it returns 0, caps won't be processed.
 *
 */

static int lids_process_flags(lids_flags_t flags)
{
	_lids_data_t *data=&lids_data[(lids_current&1)^1];

        /* LIDS switching */
#ifdef CONFIG_LIDS_ALLOW_SWITCH
	if (lids_first_time && (LIDS_FISSET(flags,LIDS_FLAGS_LIDS_ON) == 0) && (lids_load)) {
		lids_security_alert("Try to switch LIDS off at sealing time");
		return 0;
	}
	else {
		if (lids_load != LIDS_FISSET(flags,LIDS_FLAGS_LIDS_ON)) {
			lids_load=LIDS_FISSET(flags,LIDS_FLAGS_LIDS_ON);
			lids_security_alert("LIDS switched to %i",lids_load);
			if (lids_load)
				LIDS_FSET(lids_flags,LIDS_FLAGS_LIDS_ON);
			else
				LIDS_FCLR(lids_flags,LIDS_FLAGS_LIDS_ON);
		}
	}
	if (lids_first_time && (LIDS_FISSET(flags,LIDS_FLAGS_LIDS_LOCAL_ON) == 0) && (lids_load)) {
		lids_security_alert("Try to switch LIDS locally off at sealing time");
		return 0;
	}
	else {
		if (lids_local_on != LIDS_FISSET(flags,LIDS_FLAGS_LIDS_LOCAL_ON)) {
			lids_local_on=LIDS_FISSET(flags,LIDS_FLAGS_LIDS_LOCAL_ON);
			lids_security_alert("LIDS locally switched to %i",lids_local_on);
			if (lids_local_on) {
				LIDS_FSET(lids_flags,LIDS_FLAGS_LIDS_LOCAL_ON);
			}
			else {
				lids_local_pid=current->p_pptr->pid;
				if (lids_local_pid == 1) {  /* this doesn't apply to init */
					printk("Can't give local lids deactivation to init!!\n");
					LIDS_FSET(lids_flags,LIDS_FLAGS_LIDS_LOCAL_ON);
					lids_local_on = 1;
				}
				else
					LIDS_FCLR(lids_flags,LIDS_FLAGS_LIDS_LOCAL_ON); /* should not be necessary */

			}
		}
	}
#else
	if  (LIDS_FISSET(flags,LIDS_FLAGS_LIDS_ON) != lids_load) {
		lids_security_alert("Try to switch LIDS off (feature disabled)");
		return 0;
	}
#endif

	/* Config file reload */
	if (LIDS_FISSET(flags,LIDS_FLAGS_RELOAD_CONF)) {
#ifdef CONFIG_LIDS_RELOAD_CONF
		if((!lids_local_on&&lids_local_pid==current->p_pptr->pid)||!lids_load)
		{
		  data->last_secure=0;
		  data->last_s_acl=0;
		  data->last_o_acl=0;
		  lids_init();
			/* cap_bset=locks.cap_bset;*/
		   cap_bset = lids_cap_val;
		  lids_security_alert("Config. file reloaded");
		}
		else lids_security_alert("Security enabled, config. file NOT reloaded");
#else
		lids_security_alert("Try to reload config. file (feature disabled)");
#endif 
	}
	return 1;
}


/***********************************************************************
 *
 * The one which set/get security features
 *
 */

static int number_failed=0;
static int wait_after_fail=0;
static struct timer_list fail_timer;

/* called by timer */
static void reenable_sysctl(unsigned long user_data) 
{
        number_failed=0;
        wait_after_fail=0;
}

int lids_proc_locks_sysctl(ctl_table *table, int write, struct file *filp,
			    void *buffer, size_t *lenp, int conv, int op) 
{
	lids_locks_t locks;
	byte   hashcode[RMDsize/8];
	char rmd160sig[LIDS_PW_LEN+10];
	int i;
	
	/* first: check the terminal and the program which access the sysctl */

#ifndef CONFIG_LIDS_REMOTE_SWITCH
	if (current->tty && (current->tty->driver.type != 2) ) {
		lids_security_alert("Try to %s locks sysctl (unauthorized terminal)",
				    write ? "write" : "read");
		return -EPERM;
	}
#endif

#ifndef CONFIG_LIDS_ALLOW_ANY_PROG_SWITCH
	{
		struct vm_area_struct * vma;
		unsigned long ino;
		dev_t dev;
		
		ino=-1;
		dev=-1;
		if (current->mm) {
			vma = current->mm->mmap;
			while (vma) {
				if ((vma->vm_flags & VM_EXECUTABLE) && 
				    vma->vm_file && vma->vm_file->f_dentry && 
				    vma->vm_file->f_dentry->d_inode) {
					ino=vma->vm_file->f_dentry->d_inode->i_ino;
					dev=vma->vm_file->f_dentry->d_inode->i_dev;
					break;
				}
				vma = vma->vm_next;
			}
		}
		if ( (ino != lidsadm.ino) || (dev != lidsadm.dev) ) {
			lids_security_alert("Try to %s locks sysctl (unauthorised program)",
					    write ? "write" : "read");
			return -EPERM;
		}
	}
#endif
	/* second: check wether it is not a timeout period after two many failed attempts */

	if (wait_after_fail) {
		lids_security_alert("Try to %s locks sysctl during timeout",write ? "write" : "read");
		return -EPERM;
	}
	
	if (write) {
		/* Third : check what is submitted (size, magics, passwd) */
		if (*lenp != sizeof(lids_locks_t)) {
			lids_security_alert("Try to feed locks sysctl with garbage");
			return -EINVAL;
		}
		if (copy_from_user(&locks,buffer,sizeof(lids_locks_t)))
			return -EFAULT;
		if ((locks.magic1 != LIDS_MAGIC_1) || (locks.magic2 != LIDS_MAGIC_2) ||
		    (locks.magic3 != LIDS_MAGIC_3) || (locks.magic4 != LIDS_MAGIC_4)) {
			memset((char *)locks.passwd,'\0',sizeof(passwd_t));
			lids_security_alert("Try to feed locks sysctl bad magic numbers");
			return -EINVAL;
		}
		locks.passwd[sizeof(passwd_t)-1]=0; /* We don't take the risk */
		rmd160sig[0]=0;
		
#ifdef CONFIG_LIDS_ALLOW_SWITCH
		if ((!lids_first_time) || (locks.passwd[0])) {
			RMD((byte *)locks.passwd,hashcode);
			memset((char *)locks.passwd,'\0',sizeof(passwd_t));
			for (i=0; i<RMDsize/8; i++)
				sprintf(rmd160sig+2*i,"%02x", hashcode[i]);
		}
		if ( ((lids_first_time) && (!locks.passwd[0])) ||
		     (!strncmp(rmd160sig,lids_pw,LIDS_PW_LEN)) ) {
#else
		if ((lids_first_time) && (!locks.passwd[0])) {
#endif
			/* access granted ! */
			number_failed=0;
#ifdef CONFIG_LIDS_HANGUP
			if(current->tty!=NULL)
				set_bit(TTY_LIDS_NO_HANGUP, &(current->tty->flags));
#endif
			if (lids_process_flags(locks.flags)) {
				/* Seal the kernel,we can change the cap_set here */
				if(lids_first_time || LIDS_FISSET(locks.flags,LIDS_FLAGS_RELOAD_CONF))
					cap_bset = lids_cap_val;
				else 
					cap_bset = locks.cap_bset;
				lids_last_pid=0;
				if(!cap_raised(cap_bset,CAP_INIT_KILL)) {
					init_no_kill_pid();
				}
				lids_security_alert("Changed: cap_bset=0x%x lids_flags=0x%x ",cap_t(cap_bset),lids_flags);
			}
			lids_first_time=0;
		}
		else {
#ifdef CONFIG_LIDS_HANGUP
			if(current->tty!=NULL)
				clear_bit(TTY_LIDS_NO_HANGUP, &(current->tty->flags));
#endif
#ifdef CONFIG_LIDS_ALLOW_SWITCH
			number_failed++;
			lids_security_alert("Give incorrect password (try #%d) with caps=0x%x and flags=0x%x", number_failed,cap_t(locks.cap_bset),locks.flags);
			if (number_failed >= CONFIG_LIDS_MAX_TRY) {
				wait_after_fail=1;
				init_timer(&fail_timer);
				fail_timer.function=reenable_sysctl;
				fail_timer.data=(unsigned long)NULL;
				fail_timer.expires=jiffies+CONFIG_LIDS_TTW_FAIL*HZ;
				add_timer(&fail_timer);
			}
#else
			lids_security_alert("Try %d to switch caps/flags with caps=0x%x and flags=0x%x (feature disabled)",
					    number_failed,cap_t(locks.cap_bset),locks.flags);
#endif
		}
	}
	else {
		locks.cap_bset=cap_bset;
		locks.flags=lids_flags;
		if (*lenp < sizeof(lids_locks_t)) 
			return -EINVAL;
		if (copy_to_user(buffer,&locks,sizeof(lids_locks_t)))
			return -EFAULT;
	}
       return 0;
}

